const abind = require('abind');
const stackTrace = require('stack-trace');
const stringify = require('json-stringify-safe');
//
import Utils from "./../API/IKUtils";
//
export const LOG_LEVELS = { DEBUG: 0, INFO: 1, WARN: 2, ERROR: 3 };
const LOG_STRINGS = ['DEBUG', 'INFO', 'WARN', 'ERROR'];
const PURE_CONSOLE = (console.flushLogs ? console.origin : console);
const DEFAULT_LOG_FUNCTION = PURE_CONSOLE.log.bind(PURE_CONSOLE);
//
const blacklist = ['password','phonenumber'/*,'code'*/,'resetCode','recaptchaToken','token','mfa','REFRESH_TOKEN','SECRET_HASH','SecretHash','AccessToken','UserCode','paymentMethodNonce'];
//
export default class IKLogger {
  constructor(_config, _LOG_LEVEL, transactionID) {
    abind(this);
    //
    if (!_LOG_LEVEL && !(_config || {}).level) _LOG_LEVEL = LOG_LEVELS.DEBUG;
    else {
      _LOG_LEVEL = LOG_STRINGS.indexOf((_config || {}).level || _LOG_LEVEL);
      if (_LOG_LEVEL == -1) _LOG_LEVEL = LOG_LEVELS.DEBUG;
    }
    //
    this.origin = PURE_CONSOLE;
    this._LOG_LEVEL = _LOG_LEVEL;
    this._config = _config || {};
    this._transactionID = transactionID;
    this._filterBlacklist = this._config.sensitiveFilteringKeywords || blacklist;
    //
    this._setupBindings();
    //
    this.log("Using logger with level: " + LOG_STRINGS[this._LOG_LEVEL]);
    this.debug("logger config: ", this._config);
  }

  //Public
  debug(...args) { this._processLog(false, LOG_LEVELS.DEBUG, args); }
  log(...args) { this._processLog(false, LOG_LEVELS.INFO, args); }
  info(...args) { this._processLog(false, LOG_LEVELS.INFO, args); }
  warning(...args) { this._processLog(false, LOG_LEVELS.WARN, args); }
  warn(...args) { this._processLog(false, LOG_LEVELS.WARN, args); }
  error(...args) { this._processLog(false, LOG_LEVELS.ERROR, args); }
  exception(exception, ...args) { this._exception(false, exception, args); }
  _exception(isRaw ,exception, ...args) {
    //format message
    let msg = [];
    //push exeception
    msg.push(exception + " -");
    //get args
    for (let arg of args) if (arg != exception) msg.push(arg);
    msg.push(exception.stack); //push exeception stack at the end
    //push into logs stack
    this._pushLog(LOG_LEVELS.ERROR, this._formattedLog(LOG_LEVELS.ERROR, msg, this._callerName(3), isRaw));
  }
  //
  cleanUp() { }
  async flushLogs() {}


  //initialization
  _setupBindings() {
    global.console = {
      debug: (...args) => this._processLog(true, LOG_LEVELS.DEBUG, args),
      log: (...args) => this._processLog(true, LOG_LEVELS.INFO, args),
      info: (...args) => this._processLog(true, LOG_LEVELS.INFO, args),
      warning: (...args) => this._processLog(true, LOG_LEVELS.WARN, args),
      warn: (...args) => this._processLog(true, LOG_LEVELS.WARN, args),
      error: (...args) => this._processLog(true, LOG_LEVELS.ERROR, args),
      exception: (exception, ...args) => this._exception(true, exception, args),
    };
  }


  /* Formatters */
  _timestamp() {
    let d = new Date();
    return [Utils.toDoubleDigit(d.getMonth()+1), Utils.toDoubleDigit(d.getDate()), d.getFullYear()].join('/') +' ' +
           [Utils.toDoubleDigit(d.getHours()), Utils.toDoubleDigit(d.getMinutes()), Utils.toDoubleDigit(d.getSeconds())].join(':');
  }
  _formattedLog(level, msg, caller, isRaw) {
    if (Utils.isOffline()) {
      return ` [${this._timestamp()} - ${LOG_STRINGS[level]}] [${caller}] ${msg.join(" ")}`;
    } else if (Utils.isHybridlessContainer() && this._transactionID) {
      return (isRaw ? '' : ` ${this._transactionID}`) + ` [${LOG_STRINGS[level]}] [${caller}] ${(msg.map(this._supressSensitiveInfo).join(" "))}`;
    } else {
      return ` [${LOG_STRINGS[level]}] [${caller}] ${msg.join(" ")}`;
    }
  }
  _callerName(index) {
    let safeIndex = Math.min(index, stackTrace.get().length);
    if (stackTrace.get()[safeIndex]) {
      let callerName = (stackTrace.get()[safeIndex] ? stackTrace.get()[safeIndex].getTypeName() : null);
      if (!callerName) {
        callerName = stackTrace.get()[safeIndex].getFileName().split("/");
        callerName = callerName.slice(callerName.indexOf("src")).join("/");
      } return callerName + ":" + stackTrace.get()[safeIndex].getLineNumber();
    } return '';
  }
  /* Helpers (core functionality) */
  _processLog(isRaw, level, args) {
    if (level < this._LOG_LEVEL) return;
    //get args
    let msg = [];
    for (let arg of args) msg.push(((typeof arg === 'object' && !(arg instanceof Error)) ? stringify(arg, null, 2) : arg));
    //push into logs stack
    // todo: improve error stack
    this._pushLog(level, this._formattedLog(level, msg, this._callerName(3), isRaw));
  }
  _pushLog(level, fMsg) {
    // push into logs stack
    // this._logs.push(fMsg);
    DEFAULT_LOG_FUNCTION.apply(PURE_CONSOLE, [fMsg]);
  }
  /* Sensitive information handling */
  _supressSensitiveInfo(value) {
    //realy false
    if (this._config.enableSensitiveFiltering !== undefined && (!this._config.enableSensitiveFiltering || this._config.enableSensitiveFiltering == 'false')) return value;
    if (typeof value == "string") {
      //content based replacement
      blacklist.forEach((f) => {
        const match = blacklist.find((f) => value.toLowerCase().includes(f.toLowerCase()));
        if (match) value = '**SUPRESSED_SENSITIVE_DATA**';
      });
      return value;
    } else if (typeof value == 'object') {
      //key based replacement
      Object.keys(value).forEach(function (elt, i, array) {
        const match = blacklist.find((f) => elt.toLowerCase().includes(f.toLowerCase()));
        if (match) value[elt] = '**SUPRESSED_SENSITIVE_DATA**';
        else value[elt] = value[elt];
      });
      return value;
    } else if (Array.isArray(value)) return value.map(v=>this._supressSensitiveInfo(v));
    return value;
	}
}
