import RTClient from '@ikonintegration/mod-resources-tracker-client';
//
import Utils from "../API/IKUtils";
//
const fs = require('fs');
//
export default class IKExecutionTracker {
  constructor(config, context, db) {
    this.context = context;
    this.db = db;
    this.config = config;
    this.startedOn = Date.now();
    this.drainedExecution = false;
    if (config) this.client = this._getClient(config);
    this.claimers = {};
  }
  /* Public */
  appendClaimer(value, key) { this.claimers[value] = key; }
  appendEvent(event) { this.client.appendEvent(event); }
  //
  resetExecutionTrace() { 
    this.startedOn = Date.now(); 
    this.drainedExecution = false;
  }
  async quickDrain() {
    try {
      //Check for client but claimers
      if (!this.client) return; //disabled
      //drain stats (drain will filter for events with claimers only)
      await this.client.drain(this.claimers);
    } catch (e) { console.error(`Error while draining tracker events: ${e}`); }
  }
  async trackExecAndDrain() {
    try {
      //Check 
      if (this.drainedExecution) return;
      this.drainedExecution = true;
      //Check for client and claimers
      if (!this.client) return; //disabled
      //detect execution and log (ECS or lambda)
      this._detectExecutionAndLog();
      //detect DB and log (DDB only for now)
      this._collectDDBStats();
      //drain stats (drain will filter for events with claimers only)
      await this.client.drain(this.claimers);
    } catch (e) { console.error(`Error while tracking execution: ${e}`); }
  }
  /* private tracker */
  _detectExecutionAndLog() {
    //Is running on container?
    if (Utils.isHybridlessContainer()) {
      //ECS, get metadata info and log if applicable and succeeded on it
      const context = this._getECSMetadataObject();
      // console.debug('Context', context);
      if (context && context.ContainerInstanceARN) {
        const extraDetails = { startTime: this.startedOn, endTime: Date.now(), taskConfig: { cpu: process.env.CPU, memory: process.env.MEMORY, taskName: context.TaskDefinitionFamily } };
        const event = RTClient.newECSExecutionEvent(context.TaskARN, context.ContainerInstanceARN, extraDetails);
        this.appendEvent(event);
      }
    } else {
      const event = RTClient.newLambdaInvocationEvent(this.context ? this.context.awsRequestId : null);
      this.appendEvent(event); //lambda logging
    }
  }
  _collectDDBStats() {
    const stats = this.db ? this.db.getTrackerStats() : null;
    if (stats) this.client.appendEvent(RTClient.newDynamoEvent(stats.tableName, stats.readUnits, stats.writeUnits));
  }
  /* Private */
  _getClient(config) {
    return new RTClient({ busName: config.busName, region: config.region, source: config.appName });
  }
  /* ECS helper */
  _getECSMetadataObject() {
    try { return JSON.parse(fs.readFileSync(process.env.ECS_CONTAINER_METADATA_FILE)); }
    catch (e) { console.error('Unable to retrieve metadata object: ' + e); }
    return null;
  }
}
